"""Core algorithm library.

And yes, I know I shouldn't use Python for computations like these.
"""
import numpy as np

import kontrol.logger


def bisection_method(
        func, x1, x2, val, tol=1e-6, maxiter=9999, log_center=False,
        **kwargs):
    """Bisection method for finding argument of a scalar function.

    This algorithm use bisection method to find
    the argument such that a scalar function ``func(x)``
    outputs a desired target value ``val``.
    Note that in a typical root finding problem, ``val``
    is just 0.
    The target function value must exist within the
    two endponts ``x1`` and ``x2``.
    In addition ``func(x1) - val`` and ``func(x2) - val``
    must have different signs (intermediate value theorem)
    or else the algorithm won't work.
    The algorithm exits when maximum iteration is reached.

    Parameters
    ----------
    func : callable(x: float) -> float
        The function.
    x1 : float
        The first endpoint.
    x2 : float
        The second endpoint. 
    val : float
        The target value.
    ftol : float, optional,
        The tolerance of the function value for
        satisfying convergence condition.
        Defaults to 1e-6.
    maxiter : ``int``, optional
        Maximum number of iteration before termination.
        Defaults to 9999.
    log_center : bool, optional
        Use the logarithmic center of the endpoints to calculate
        the next endpoint instead of the using the linear center.
        Defaults to ``False``.
    **kwargs
        Keyword arguments passed to ``func(x, **kwargs)``.

    Returns
    -------
    xm : float or None
        The solutuon where ``func(xm)=val``.
        If maxiter is reached, returns ``None``

    Raises
    ------
    ValueError
        If ``func(x1) - val`` and ``func(x2) - val`` don't have
        different signs.

    Notes
    -----
    The termination condition is ``abs(func(xm)-val) < tol``.
    """
    func_x1 = func(x1, **kwargs)
    func_x2 = func(x2, **kwargs)
    if np.sign(func_x1 - val) == np.sign(func_x2 - val):
        raise ValueError("func(x1)-val and func(x2)-val must have different"
                         " signs.")

    for _ in range(maxiter):
        if log_center:
            xm = 10**((np.log10(x1) + np.log10(x2))/2)
        else:
            xm = (x1+x2)/2
        func_val = func(xm, **kwargs)
        if abs((func_val-val)) < tol:
            return xm
        else:
            func_x1 = func(x1, **kwargs)
            func_x2 = func(x2, **kwargs)
            if np.sign(func_x1 - val) == np.sign(func_val - val):
                x1 = xm
            else:
                x2 = xm
    kontrol.logger.logger.error("Maximum number of iteration reached")
    return None


    



# from . import frequency_series
# from .frequency_series import FrequencySeries


# from .frequency_series.frequency_series import FrequencySeries
from .frequency_series.frequency_series import FrequencySeries
from .complementary_filter.complementary_filter import ComplementaryFilter
from .transfer_function import *
from .sensact.matrix  import Matrix, SensingMatrix
from .sensact.optical_lever import (
    OpticalLeverSensingMatrix, HorizontalOpticalLeverSensingMatrix,
    VerticalOpticalLeverSensingMatrix)
from .core import spectral, foton
from . import dmd


## for ad hoc/optional packages
import importlib
ezca_spec = importlib.util.find_spec("ezca")
ezca_exist = ezca_spec is not None
if ezca_exist:
    from .ezca import Ezca



import logging

logger = logging.getLogger("Kontrol")

logger.setLevel(logging.DEBUG)

# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

# create formatter
formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)-8s: %(message)s', datefmt='%H:%M')

# add formatter to ch
ch.setFormatter(formatter)

# add ch to logger
logger.addHandler(ch)
logger.propagate = False



"""Base class for curve fitting
"""
import numpy as np


class CurveFit:
    """Base class for curve fitting

    Parameters
    ----------
    xdata : array or None, optional
        The independent variable data / data on the x axis.
        Defaults to None.
    ydata : array or None, optional
        The dependent variable data / data on the y axis.
        Defaults to None.
    model : func(x: array, args: array, **kwargs) -> array, or None, optional
        The model used to fit the data.
        ``args`` in model is an array of parameters that
        define the model.
        Defaults to None
    model_kwargs : dict or None, optional
        Keyword arguments passed to the model.
        Defaults to None.
    cost : kontrol.curvefit.Cost or func(args, model, xdata, ydata) -> array
        Cost function.
    xdata: array, ydata: array) -> float
        The cost function to be used to fit the data.
        First argument is a list of parameters that will be passed to
        the model.
        This must be pickleable if multiprocessing is to be used.
        Defaults to None.
    optimizer : func(func, **kwargs) -> OptimizeResult, or None, optional
        The optimization algorithm use for minimizing the cost function.
    optimizer_kwargs : dict or None, optional
        Keyword arguments passed to the optimizer function.
        Defaults to None.
    """
    def __init__(self, xdata=None, ydata=None,
                 model=None, model_kwargs=None,
                 cost=None, optimizer=None, optimizer_kwargs=None):
        """Constructor

        Parameters
        ----------
        xdata : array or None, optional
            The independent variable data / data on the x axis.
            Defaults to None.
        ydata : array or None, optional
            The dependent variable data / data on the y axis.
            Defaults to None.
        model : func(x: array, args: array, **kwargs) -> array, or None, optional
            The model used to fit the data.
            ``args`` in model is an array of parameters that
            define the model.
            Defaults to None
        model_kwargs : dict or None, optional
            Keyword arguments passed to the model.
            Defaults to None.
        cost : kontrol.curvefit.Cost or func(args, model, xdata, ydata) -> array
            Cost function.
            The cost function to be used to fit the data.
            First argument is a list of parameters that will be passed to
            the model.
            This must be pickleable if multiprocessing is to be used.
            Defaults to None.
        optimizer : func(func, **kwargs) -> OptimizeResult, or None, optional
            The optimization algorithm use for minimizing the cost function.
        optimizer_kwargs : dict or None, optional
            Keyword arguments passed to the optimizer function.
            Defaults to None.
        """
        self._xdata = None
        self._ydata = None
        self._model = None
        self._model_kwargs = None
        self._cost = None
        self._optimizer = None
        self._optimizer_kwargs = None
        self._optimized_args = None
        self._optimize_result = None
        self.xdata = xdata
        self.ydata = ydata
        self.model = model
        self.model_kwargs = model_kwargs
        self.cost = cost
        self.optimizer = optimizer
        self.optimizer_kwargs = optimizer_kwargs
        self.optimized_args = None
        self.optimize_result = None

    def prefit(self):
        """Something to do before fitting."""
        pass

    def fit(self, model_kwargs=None, cost_kwargs=None, optimizer_kwargs=None):
        """Fit the data

        Sets self.optimized_args and self.optimize_result.

        Parameters
        ----------
        model_kwargs : dict or None, optional
            Overriding keyword arguments in self.model_kwargs.
            Defaults to None.
        cost_kwargs : dict or None, optional
            Overriding keyword arguments in self.cost_kwargs.
            Defaults to None.
        optimizer_kwargs : dict or None, optional
            Overriding keyword argumetns in self.optimizer_kwargs.
            Defaults to None.

        Returns
        -------
        scipy.optimizer.OptimizeResult
        """
        self.prefit()
        if model_kwargs is None:
            model_kwargs = self.model_kwargs
        if optimizer_kwargs is None:
            optimizer_kwargs = self.optimizer_kwargs

        cost = self.cost
        model = self.model
        xdata = self.xdata
        ydata = self.ydata
        optimizer = self.optimizer
        if (cost is None or model is None or xdata is None or ydata is None or
                optimizer is None):
            raise TypeError("Cost, model, xdata, ydata, and optimizer must"
                            " be specified before fitting.")

        res = optimizer(
            cost, args=(model, xdata, ydata, model_kwargs), **optimizer_kwargs)
        self.optimize_result = res
        self.optimized_args = res.x
        self.model.args = self.optimized_args
        return res

    @property
    def xdata(self):
        """The independent variable data."""
        return self._xdata

    @xdata.setter
    def xdata(self, _xdata):
        """xdata setter"""
        self._xdata = _xdata

    @property
    def ydata(self):
        """The dependent variable data"""
        return self._ydata

    @ydata.setter
    def ydata(self, _ydata):
        """ydata setter"""
        self._ydata = _ydata

    @property
    def model(self):
        """The model used to fit the data."""
        return self._model

    @model.setter
    def model(self, _model):
        """model setter"""
        self._model = _model

    @property
    def model_kwargs(self):
        """Keyword arguments passed to the model."""
        return self._model_kwargs

    @model_kwargs.setter
    def model_kwargs(self, _model_kwargs):
        """model_kwargs setter"""
        if _model_kwargs is None:
            self._model_kwargs = {}
        else:
            self._model_kwargs = _model_kwargs

    @property
    def cost(self):
        """The cost function to be used to fit the data."""
        return self._cost

    @cost.setter
    def cost(self, _cost):
        """cost setter"""
        self._cost = _cost

    @property
    def optimizer(self):
        """The optimizer function to be used to fit the data."""
        return self._optimizer

    @optimizer.setter
    def optimizer(self, _optimizer):
        """optimizer setter"""
        self._optimizer = _optimizer

    @property
    def optimizer_kwargs(self):
        """Keyword arguments passed to the optimizer function."""
        return self._optimizer_kwargs

    @optimizer_kwargs.setter
    def optimizer_kwargs(self, _optimizer_kwargs):
        """optimizer_kwargs setter"""
        if _optimizer_kwargs is None:
            self._optimizer_kwargs = {}
        else:
            self._optimizer_kwargs = _optimizer_kwargs

    @property
    def optimized_args(self):
        """The optimized arguments"""
        return self._optimized_args

    @optimized_args.setter
    def optimized_args(self, _optimized_args):
        """optimized_args setter"""
        self._optimized_args = _optimized_args

    @property
    def optimize_result(self):
        """Optimization Result"""
        return self._optimize_result

    @optimize_result.setter
    def optimize_result(self, _optimize_result):
        """optimize_result setter"""
        self._optimize_result = _optimize_result

    @property
    def yfit(self):
        """The fitted y values"""
        if self.optimized_args is None:
            return None
        else:
            return self.model(
                self.xdata, self.optimized_args, **self.model_kwargs)
